from typing import Callable, Any, Optional, Union

import pika
from marshmallow import Schema
from pycapo import CapoConfig
from workspaces.json import WorkflowEventSchema


CONNECTION = None
CHANNEL = None
CONFIG = None


class Channel:
    def __init__(self, schema: Schema):
        self.schema = schema

    def connect(self, **kwargs: Union[int, str]) -> pika.BlockingConnection:
        """
        Initialize connection to AMQP server given a CAPO profile
        :return: Established connection
        """
        global CONNECTION
        global CHANNEL
        global CONFIG

        if not CONNECTION:
            CONFIG = CapoConfig(
                profile=kwargs.get('profile', None)
            ).settings('edu.nrao.archive.configuration.AmqpServer')

            connection_parameters = pika.ConnectionParameters(
                host=kwargs.get('hostname', CONFIG.hostname),
                port=int(kwargs.get('port', CONFIG.port)),
                # FIXME: Copied from events. Do these args need to be cast to int?
                connection_attempts=kwargs.get('connection_attempts', 5),
                socket_timeout=kwargs.get('socket_timeout', 5000),
                retry_delay=kwargs.get('retry_delay', 500),
                #
                credentials=pika.PlainCredentials(
                    username=kwargs.get('username', CONFIG.username),
                    password=kwargs.get('password', CONFIG.password)
                )
            )
            CONNECTION = pika.BlockingConnection(connection_parameters)
            CHANNEL = CONNECTION.channel()
            CHANNEL.exchange_declare(
                # FIXME: CAPO config does not currently support 'exchange' value
                exchange=kwargs.get('exchange', CONFIG.exchange),
                exchange_type='topic',
                durable=True,
                auto_delete=False
            )

    def close(self):
        if CONNECTION:
            CONNECTION.close()

    def send(self, event: WorkflowEventSchema):
        rendered = self.schema.dump(event)
        routing_key = f'{rendered["job_name"]}.{rendered["job_id"]}.{rendered["event_type"]}'
        CHANNEL.basic_publish(routing_key=routing_key, body=str(rendered))

    def listen(self, callback: Optional[Callable], pattern: str = '#', **kwargs: Union[str, bool]):
        """
        Establishes queue and binds it to a given channel and consumes messages matching the
        routing key from given exchange
        :param callback: Callback function for when a message is consumed
        :param pattern: Pattern to be used as routing key
        Optional keyword arguments
        :param exchange: AMQP exchange name
        :param auto_ack: If true, consumer automatically acknowledges when it consumes a message
        """
        exchange = kwargs.get('exchange', CONFIG.exchange)
        auto_ack = kwargs.get('auto_ack', False)

        def unwrapping_callback(message):
            event = self.schema.load(message)
            callback(event)
        # FIXME: ask the underlying channel to call us back
        # but only for messages matching pattern

        queue = CHANNEL.queue_declare(queue='', exclusive=True).method.queue
        CHANNEL.queue_bind(queue=queue, exchange=exchange, routing_key=pattern)
        CHANNEL.basic_consume(queue=queue, on_message_callback=callback, auto_ack=auto_ack)
        CHANNEL.start_consuming()

    def __enter__(self, **kwargs: Union[int, str]):
        """
        Keyword arguments for the AMQP connection. These do not need to be specified:
        :param: hostname: Hostname to connect to
        :param: port: Port to connect to
        :param: connection_attempts: Number of connection attempts to try
        :param: socket_timeout: Time to wait for a socket to connect
        :param: retry_delay: Time to wait between retrying the connection
        :param: username: Username to connect to as
        :param: password: Password to use when connecting
        :param: exchange: Exchange to use when connection
        """
        self.connect(**kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


# Predefined event channels for ease of use
workflow_events = Channel(WorkflowEventSchema())
